@@ -685,10 +685,9 @@ static void print_raid_stripe_key(struct extent_buffer *eb,
 
 	printf("\t\t\tencoding: %s\n", stripe_encoding_name(encoding));
 	for (int i = 0; i < num_stripes; i++)
-		printf("\t\t\tstripe %d devid %llu physical %llu length %llu\n", i,
+		printf("\t\t\tstripe %d devid %llu physical %llu\n", i,
 		       (unsigned long long)btrfs_raid_stride_devid_nr(eb, stripe, i),
-		       (unsigned long long)btrfs_raid_stride_offset_nr(eb, stripe, i),
-		       (unsigned long long)btrfs_raid_stride_length_nr(eb, stripe, i));
+		       (unsigned long long)btrfs_raid_stride_offset_nr(eb, stripe, i));
 }
 
 void print_key_type(FILE *stream, u64 objectid, u8 type)
